
#include <iostream>
#include <fstream>
#include <vector>
//#include <set>
#include <string>
#include <cstdio>
#include <cctype>
#include <algorithm>
//#include <functional>
#include "Project.h"

/*******************************************************
*
*	Public member functions 
*
*******************************************************/


// Construct a Project by reading a text file
// which specifies the names, costs, 
// and prerequisites of the activities.
//
Project::Project(const string &filename) {
	readfile(filename);
}

// Display the nodes (activities) and their costs
//
void Project::displayNodes() {
	cout << "There are " << nodes.size() << " activities --\n";
	for (node_p p = nodes.begin(); p!=nodes.end(); ++p) {
		cout << *p << "\n";
	}
	cout << endl;
}

// Display the links (a.k.a. arcs or prereqs)
//
void Project::displayLinks() {
	cout << "There are " << links.size() << " prerequisites (from to) --\n";
	for (link_p p = links.begin(); p!=links.end(); p++) {
		cout << p->first.getName() << " " << p->second.getName() << "\n";
		//cout << p->second << " " << p->first << "\n";
	}
	cout << endl;
}

// Display the paths
//
void Project::displayPaths() {
	cout << "There are " << paths.size() << " paths -- \n";
	for (int i=0; i<paths.size(); i++) {
		cout << paths[i] << endl;
	}
	cout << endl;
}

// Display the critical paths.
// Assume the paths have been initialized and sorted.
//
void Project::displayCriticalPaths() {

	if (paths.size() == 0) {
		cout << "There are no critical paths. \n";
		return;
	}

	// Use the STL count algorithm.
	// This works because class Path has an operator==()
	//
	int ncritical = count(paths.begin(), paths.end(), paths[0]);

	int cost = paths[0].getCost();
	cout << "There are " << ncritical << " critical paths -- \n";
	for (int i=0; i<paths.size(); i++) {
		if (paths[i].getCost() == cost)
			cout << paths[i] << endl;
		else
			break;
	}
	cout << endl;
}

/*******************************************************
*
*	Private member functions
*
*******************************************************/

// Initialize the Project by reading a file
//
bool Project::readfile(const string &filename) {
	status = false;

	vector<pair<string,string> > name_pairs;

	ifstream fs(filename.c_str());
	if (!fs)
		error("Can't open file '" + filename + "'");

	enum input_state { NONE, COST, PREREQ };
	input_state istate = COST;
	string line;
	for (int lineno=1; !fs.eof(); ++lineno) {
		getline(fs, line);
		if (fs.bad())
			error("Error reading file '" + filename + "'");
		//cout << "line = '" << line << "'" << endl;
		stripSpaces(line);

		if (line.length() == 0 || line[0] == '#') {
			; // skip blank lines and comment lines
		} else if (line == "costs:") {
			istate = COST;
		} else if (line == "prereqs:") {
			istate = PREREQ;
		} else {
			vector<string> v = split(line);
			switch (istate) {
			case COST :
				if (v.size() != 2)
					format_error("expected: nodename cost", lineno);
				createNode(v[0], atoi(v[1].c_str()));
				break;
			case PREREQ : 
				if (v.size() != 2)
					format_error("expected: nodename1 nodename2", lineno);
				name_pairs.push_back(make_pair(v[0],v[1]));
				break;
			}
		}
	}
	fs.close(); // would be done by destructor

#if 0
	// sort the list of nodes
	//
	sort(nodes.begin(), nodes.end());
#endif

	// make the table of prerequisites
	//
	for (int i = 0 ; i<name_pairs.size(); ++i) {
		string name1 = name_pairs[i].first;
		string name2 = name_pairs[i].second;
		if (name1 == name2)
			error("node '" + name1 + "' is defined as its own prerequisite");
		Node *n1 = findNode(name1);
		if (!n1) 
			error("cost of prerequisite node '" + name1 +"' not specified");
		Node *n2 = findNode(name2);
		if (!n2) 
			error("cost of node '" + name2 +"' not specified");
		links.insert(make_pair(*n1, *n2));
		n2->setStartNode(false);  // has a prereq, can't be start node.
	}

	// Calculate the paths
	//
	makePaths();
	sort(paths.begin(), paths.end());
	reverse(paths.begin(), paths.end());

	//
	status = true;
	return status;
}

// Create a Node, add to the list for this project
//
void Project::createNode(const string &name, int cost) {
	Node n(name, cost);
	nodes.push_back(n);
}

// Find a Node, given its name.
// Returns 0 if not found.
//
Node *Project::findNode(const string &name) {
	Node n(name);
	node_p p = find(nodes.begin(), nodes.end(), n);
	return (p==nodes.end()) ? 0 : &(*p);
}

// Destructor
//
Project::~Project() {
}

// Calculate all of the Paths.
//
// For each starting node (i.e. node without prereqs),
// find all of the path that begin with that node,
// and add them to the list of paths for this project.
//
void Project::makePaths() {
	for (node_p p=nodes.begin(); p!=nodes.end(); ++p) { // for each node
		if (p->isStartNode()) {	// if no prereqs
			seen_type empty;
			vector<Path> v = getPathsFromNode(*p, empty);
			for (vector<Path>::iterator vi=v.begin(); vi!=v.end(); ++vi) 
			{
				vi->calcCost();		// finalize
				addPath(*vi);		// add path to project
			}
		}
	}
}

// Return the paths starting at a given node.
//
// This is a recursive function.
// The parameter 'seen' contains the Nodes which have
// been traversed on the current path.
//
vector<Path> Project::getPathsFromNode(Node &nd, seen_type &seen) 
{
	// check for a loop in the prereqs
	//
	seen_type::iterator si = find(seen.begin(), seen.end(), nd);
	if (si != seen.end()) {
		string msg = "circular path at node '" + nd.getName() + "'";
		msg += "\n(Nodes seen: ";
		for (si=seen.begin(); si != seen.end(); ++si)
			msg += si->getName() + ", ";
		msg += nd.getName() + ")";
		error(msg);
	}

	vector<Path> result;

	pair<link_p, link_p> range;
	range = links.equal_range(nd);
	//
	if (range.first == range.second) {
		result.push_back(Path(nd));		// no goals; path of one node
	} else {
		// add to the list of nodes traversed
		// make copy; don't modify caller's copy
		seen_type seen2(seen);
		seen2.push_back(nd);

		// for each goal of this node ...
		for (link_p l=range.first; l!=range.second; ++l) {
			Node goal = l->second;
			vector<Path> subPaths = getPathsFromNode(goal, seen2);
			for (vector<Path>::iterator vi=subPaths.begin(); 
				vi!=subPaths.end(); ++vi) 
			{
				vi->push_front(nd);		// add this node to front
				result.push_back(*vi);	// add this path to results
			}
		}
	}
	return result;
}


// Add to the list of Paths
//
void Project::addPath(Path &p) {
	paths.push_back(p);
}

// Throw an exception
//
void Project::error(const string &msg) const {
	throw msg;
}

// Throw a exception because of a syntax error in the input file
//
void Project::format_error(const string &msg, int lineno) const {

	// Append the string representation of int.
	// Not sure if ostringstream or ostrstream is implemented,
	// so use the C method.
	char buf[8];
	sprintf(buf, "%d", lineno);
    throw string("bad format, line ") + buf + "; " + msg;
}

// strip leading and trailing whitespace
//
string & Project::stripSpaces(string & s) const {
	for (string::iterator p = s.begin(); p != s.end(); ) {
		if (isspace(*p))
			p = s.erase(p);
		else break;
	}
	for (string::iterator p=s.end(); s.length()>0;) {
		--p;
		if (isspace(*p))
			s.erase(p);
		else break;
	}

	return s;
}

// split a string at whitespace
//
vector<string> Project::split(const string &s) const {
	string ws = "\t\n\v\f\r ";
	vector<string> v;
	string t(s);
	while (t.length() > 0) {

		// strip leading spaces
		while (t.length() > 0 && ws.find(t[0]) != string::npos) {
			t.erase(0,1);
		}

		if (t.length() == 0)
			break;

		// find first whitespace (after non-whitespace)
		int i = t.find_first_of(ws);

		if (i == string::npos) {
			v.push_back(t);
			break;
		}
		v.push_back(t.substr(0,i));
		t.erase(0,i);
	}
	return v;
}

